(* Berechnung zu Sonne und Mond
   ----------------------------
   Umsetzung der Routinen aus "SunMoon.cs" von Stephan Schweig nach Delphi
    Dr. J. Rathlev, D-24222 Schwentinental (kontakt(a)rathlev-home.de)

   Info in "SunMoon.cs":
   Entfernen Sie folgende Informationen auf keinen Fall: / Do not remove following text:
   Source code based on the javascript by Arnold Barmettler, www.astronomie.info / www.CalSky.com
   based on algorithms by Peter Duffett-Smith's great and easy book
   'Practical Astronomy with your Calculator'.

   The contents of this file may be used under the terms of the
   Mozilla Public License ("MPL") or
   GNU Lesser General Public License Version 2 or later (the "LGPL")

   Software distributed under this License is distributed on an "AS IS" basis,
   WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
   the specific language governing rights and limitations under the License.

   Version 1: July 2019
*)

unit SunMoon;

interface

type
  // Tierkreiszeichen
  TStarSign =  (ssAries,       	//!< Widder
                ssTaurus,      	//!< Stier
                ssGemini,      	//!< Zwillinge
                ssCancer,      	//!< Krebs
                ssLeo, 		//!< Lwe
                ssVirgo,       	//!< Jungfrau
                ssLibra,       	//!< Waage
                ssScorpio,     	//!< Skorpion
                ssSagittarius,  //!< Schtze
                ssCapricornus,  //!< Steinbock
                ssAquarius,    	//!< Wassermann
                ssPisces);     	//!< Fische

  // Mondphasen
  TLunarPhase = (lpNewMoon,              //!< Neumond
		lpWaxingCrescentMoon,    //!< Zunehmende Sichel
                lpFirstQuarterMoon,      //!< Erstes Viertel
                lpWaxingGibbousMoon,     //!< Zunehmender Mond
                lpFullMoon,              //!< Vollmond
                lpWaningGibbousMoon,     //!< Abnehmender Mond
                lpLastQuarterMoon,       //!< Letztes Viertel
                lpWaningCrescentMoon);   //!< Abnehmende Sichel

  TTimeSpan = record
    hh,mm,ss : word;   //!< Stunden, Minuten, Sekunden
    float : double;    //!< Zeit als Fliekommazahl in Stunden
    end;

  TSunMoonData = record
    Lat,Lon : double;	             //!< Geografische Breite und Lnge in Grad
    DateTime : TDateTime;            //!< Datum und Uhrzeit
    Zone : integer;  		     //!< Zeitdifferenz zu Weltzeit (UTC) in h
    DeltaT,		             //!< deltaT in sek
    JD : double;		     //!< Julianisches Datum in Tagen seit 1.1.1900
    GMST,			     //!< Greenwich Sternzeit GMST in h
    LMST : TTime;  		     //!< Lokale Sternzeit LMST in h
    EOT : double;                    //!< Zeitgleichung in min
    SunDistance,                     //!< Entfernung der Sonne (Erdmittelpunkt) in km
    SunDistanceObserver,             //!< Entfernung der Sonne (vom Beobachter) in km
    SunLon : double;		     //!< Eklipt. Lnge der Sonne in Grad
    SunRA : TTime;		     //!< Rektaszension der Sonne in h
    SunDec,			     //!< Deklination der Sonne in Grad
    SunAz,			     //!< Azimut der Sonne in Grad
    SunAlt,			     //!< Hhe der Sonne ber Horizont in Grad
    SunDiameter : double;	     //!< Durchmesser der Sonne in ' (Bogenmimuten)
    SunAstronomicalTwilightMorning,  //!< Astronomische Morgendmmerung in h
    SunNauticalTwilightMorning,	     //!< Nautische Morgendmmerung in h
    SunCivilTwilightMorning,	     //!< Brgerliche Morgendmmerung in h
    SunRise,			     //!< Sonnenaufgang in h
    SunTransit,			     //!< Sonnenkulmination in h
    SunSet,			     //!< Sonnenuntergang in h
    SunCivilTwilightEvening,	     //!< Brgerliche Abenddmmerung in h
    SunNauticalTwilightEvening,	     //!< Nautische Abenddmmerung in h
    SunAstronomicalTwilightEvening : TTime;//!< Astronomische Abenddmmerung in h
    SunSign : TStarSign;	     //!< Tierkreiszeichen
    MoonDistance,		     //!< Entfernung des Mondes (Erdmittelpunkt) in km
    MoonDistanceObserver,	     //!< Entfernung des Mondes (vom Beobachter) in km
    MoonLon,			     //!< Eklipt. Lnge des Mondes in Grad
    MoonLat : double;		     //!< Eklipt. Breite des Mondes in Grad
    MoonRA : TTime;	             //!< Rektaszension des Mondes in h
    MoonDec,			     //!< Deklination des Mondes in Grad
    MoonAz,			     //!< Azimut des Mondes in Grad
    MoonAlt,			     //!< Hhe des Mondes ber Horizont in Grad
    MoonDiameter : double;	     //!< Durchmesser des Mondes in ' (Bogenmimuten)
    MoonRise,                        //!< Mondaufgang in h
    MoonTransit,                     //!< Mondkulmination in h
    MoonSet : TTime;		     //!< Monduntergang in h
    MoonPhaseNumber,                 //!< Mondphase
    MoonAge : double;	             //!< Mondalter in Grad
    MoonPhase : TLunarPhase;	     //!< Mondphase
    MoonSign : TStarSign;            //!< Mondzeichen
    end;

  TSunTimes = record
    SunRise,			     //!< Sonnenaufgang in h
    SunTransit,			     //!< Sonnenkulmination in h
    SunSet  : TTime;		     //!< Sonnenuntergang in h
    end;

function TimeOfDayToStr(tm : TTime; NoSeconds : boolean = false) : string;

function InitData : TSunMoonData;
function ComputeSunMoon (latitude,longitude,height : double; tm : TDateTime; aDeltaT : double; aZone : integer) : TSunMoonData;
function ComputeSunTimes  (latitude,longitude : double; tm : TDateTime; aDeltaT : double; aZone : integer) : TSunTimes;
function SunRightAscension (latitude,longitude : double; tm : TDateTime; aDeltaT : double; aZone : integer) : TTime;
function MoonPhase (latitude,longitude : double; tm : TDateTime; aDeltaT : double; aZone : integer) : TLunarPhase;
function MoonPhaseNumber (latitude,longitude : double; tm : TDateTime; aDeltaT : double; aZone : integer) : double;
function MoonAge (latitude,longitude : double; tm : TDateTime; aDeltaT : double; aZone : integer) : double;

implementation

uses System.SysUtils, System.Math, System.DateUtils;

const
  xDeg = Pi/180;
  xRad = 180/Pi;

type
  TSunMoonStg = record
    error : boolean;
    lat,lon,ra,dec,
    az,alt,radius,distance,
    rise,transit,xset,
    x,y,z,
    distanceTopocentric,
    decTopocentric,raTopocentric,anomalyMean,
    diameter,parallax,orbitLon,
    raGeocentric,decGeocentric,
    moonAge,phase : double;
    moonPhase : integer;
    civilTwilightMorning,civilTwilightEvening,
    nauticalTwilightMorning,nauticalTwilightEvening,
    astronomicalTwilightMorning,astronomicalTwilightEvening : double;
    end;
  PSunMoonStg = ^TSunMoonStg;

procedure IncFloat (var x : double; y : double);
begin
  x:=x+y;
  end;

function Mod2Pi (x : double) : double;
begin
  while x<=0 do IncFloat(x,2*Pi);
  Result:=FMod(x,2*Pi);
  end;

// Calculate Julian day - valid after -4713 November 23
// see: https://de.wikipedia.org/wiki/Julianisches_Datum
function CalcJD (day,month,year : word) : double;
var
  b : double;
begin
  if (month <= 2) then begin
    dec(year); inc(month,12);
    end;
  b:=2-Floor(year/100)+Floor(year/400);
  Result:=Floor(365.25*(year+4716))+Floor(30.6001*(month+1))+day+b-1524.5;
  end;

// Julian Date to Greenwich Mean Sidereal Time
function CalcGMST (JD : double) : double;
var
  UT,T : double;
begin
  UT:=Frac(JD-0.5)*24.0;   // UT in hours
  JD:=Floor(JD-0.5)+0.5;   // JD at 0 hours UT
  T:=(JD-2451545.0)/36525.0;
  Result:=FMod(6.697374558+T*(2400.051336+T*0.000025862)+1.002737909*UT,24.0);
  end;

// calculate equation of time
// from: http://www.nature1st.net/bogan/astro/time/jsjdetst.html
function calcEOT (JD : double) : double;
var
  T,xo,xl,xm,xe : double;
begin
  T:=(JD-2451545.0)/36525;
  xo:=(23.439291-0.001300*T)/xRad;
  xm:=(((-0.00000048*T-0.0001559)*T+35999.05030)*T+357.05291)/xRad;
  xe:=(-0.0000001236*T-0.000042037)*T+0.01670617;
  xo:=sqr(tan(xo/2));
  T:=T/10;
  xl:=((((1/15299*T+1/49931)*T+0.030320)*T+360007.6982779)*T+280.46644567)/xRad;
  Result:=4*xRad*(xo*sin(2*xl)-2*xe*sin(xm)+(4*xe*sin(xm)*cos(2*xl)-xo/2*sin(4*xl))*xo);
  end;

// Convert Greenwich Mean Sidereal Time to UT
function GMST2UT(JD,gmst : double) : double;
var
  T : double;
begin
  JD:=Floor(JD-0.5)+0.5;   // JD at 0 hours UT
  T:=(JD-2451545.0)/36525.0;
  Result:=0.9972695663*((gmst-FMod(6.697374558+T*(2400.051336+T*0.000025862),24.0)));
  end;

// Local Mean Sidereal Time, geographical longitude in radians, East is positive
function GMST2LMST(gmst,lon : double) : double;
begin
  Result:=FMod(gmst+xRad*lon/15,24.0);
  end;

// Transform ecliptical coordinates (lon/lat) to equatorial coordinates (RA/dec)
procedure Ecl2Equ(TDT : double; var coor : TSunMoonStg);
var
  T,eps,coseps,sineps,sinlon : double;
begin
  T:=(TDT - 2451545.0) / 36525.0; // Epoch 2000 January 1.5
  eps:=(23.0 + (26 + 21.45 / 60.0) / 60.0 + T * (-46.815 + T * (-0.0006 + T * 0.00181))/3600.0)*xDeg;
  coseps:=Cos(eps); sineps:=Sin(eps);
  with coor do begin
    sinlon:=sin(lon);
    ra:=mod2pi(arctan2((sinlon * coseps - tan(lat) * sineps), cos(lon)));
    dec:=arcsin(sin(lat) * coseps + cos(lat) * sineps * sinlon);
    end;
  end;

// Transform equatorial coordinates (RA/Dec) to horizonal coordinates (azimuth/altitude)
// Refraction is ignored
procedure Equ2Altaz (TDT,geolat,lmst : double; var coor : TSunMoonStg);
var
  cosdec,sindec,coslha,sinlha,coslat,sinlat,lha : double;
begin
  with coor do begin
    cosdec:=cos(dec); sindec:=sin(dec);
    lha:=lmst-ra; coslha:=cos(lha); sinlha:=sin(lha);
    coslat:=cos(geolat); sinlat:=sin(geolat);
    az:=mod2pi(arctan2(-cosdec * sinlha,sindec * coslat - cosdec * coslha * sinlat));
    alt:=arcsin(sindec * sinlat + cosdec * coslha * coslat);
    end;
  end;

// Transform geocentric equatorial coordinates (RA/Dec) to topocentric equatorial coordinates
procedure GeoEqu2TopoEqu(lmst : double; const observer : TSunMoonStg; var coor : TSunMoonStg);
var
  cosdec,sindec,coslst,sinlst,coslat,sinlat,rho,xx,xy,xz : double;
begin
  coslst:=cos(lmst);
  sinlst:=sin(lmst);
  coslat:=cos(observer.lat); // we should use geocentric latitude, not geodetic latitude
  sinlat:=sin(observer.lat);
  rho:=observer.radius; // observer-geocenter in Kilometer
  with coor do begin
    cosdec:=cos(dec);
    sindec:=sin(dec);
    xx:=distance * cosdec * cos(ra) - rho * coslat * coslst;
    xy:=distance * cosdec * sin(ra) - rho * coslat * sinlst;
    xz:=distance * sindec - rho * sinlat;
    distanceTopocentric:=sqrt(sqr(xx) + sqr(xy) + sqr(xz));
    decTopocentric:=arcsin(xz / distanceTopocentric);
    raTopocentric:=mod2pi(arctan2(xy, xx));
    end;
  end;

// Calculate cartesian from polar coordinates
function EquPolar2Cart(alon,alat,adistance : double) : TSunMoonStg;
var
  rcd : double;
begin
  rcd:=cos(alat) * adistance;
  FillChar(Result,sizeof(TSunMoonStg),0);
  with Result do begin
    x:=rcd * cos(alon);
    y:=rcd * sin(alon);
    z:=adistance * sin(alat);
    end;
  end;

// Calculate observers cartesian equatorial coordinates (x,y,z in celestial frame)
// from geodetic coordinates (longitude, latitude, height above WGS84 ellipsoid)
// Currently only used to calculate distance of a body from the observer
function Observer2EquCart(alon,alat,height,gmst : double) : TSunMoonStg;
var
  co,si,fl,u,a,b,xr,xx,xy,rotangle : double;
const
  flat = 298.257223563;        // WGS84 flatening of earth
  aearth = 6378.137;           // GRS80/WGS84 semi major axis
begin
//  FillChar(Result,sizeof(TSunMoonStg),0);
  // Calculate geocentric latitude from geodetic latitude
  co:=cos(alat);
  fl:=sqr(1.0 - 1.0 / flat);
  si:=sqr(sin(alat));
  u:=1.0 / sqrt(sqr(co) + fl * si);
  a:=aearth * u + height;
  b:=aearth * fl * u + height;
  xr:=sqrt(sqr(a) * sqr(co) + sqr(b) * si); // geocentric distance from earth center
  xx:=alon; // longitude stays the same
  xy:=arccos(a * co / xr); // geocentric latitude, rad
  if (alat < 0.0) then xy:=-xy;  // adjust sign
  Result:=EquPolar2Cart(xx,xy,xr); // convert from geocentric polar to geocentric cartesian, with regard to Greenwich
  with Result do begin
    // rotate around earth's polar axis to align coordinate system from Greenwich to vernal equinox
    xx:=x; xy:=y;
    rotangle:=gmst / 24.0 * 2.0 * Pi; // sideral time gmst given in hours. Convert to radians
    x:=xx * cos(rotangle) - xy * sin(rotangle);
    y:=xx * sin(rotangle) + xy * cos(rotangle);
    radius:=xr;
    lon:=alon;
    lat:=alat;
    end;
  end;

// Calculate coordinates for Sun
// Coordinates are accurate to about 10s (right ascension)
// and a few minutes of arc (declination)
function SunPosition(TDT,geolat,lmst : double) : TSunMoonStg;
var
  wg,diameter0,MSun,xnu : double;
const
  e = 0.016713;
  a = 149598500; // km
begin
  FillChar(Result,sizeof(TSunMoonStg),0);
  wg:=282.768422 * xDEG;
  diameter0:=0.533128 * xDEG; // angular diameter of Moon at a distance
  MSun:=2*Pi / 365.242191 * (TDT - 2447891.5) + (279.403303 * xDEG) - wg;
  xnu:=MSun + 2 * e * sin(MSun);
  with Result do begin
    lon:=mod2pi(xnu + wg);
    lat:=0;
    anomalyMean:=MSun;
    distance:=(1 - sqr(e)) / (1 + e * cos(xnu)); // distance in astronomical units
    diameter:=diameter0 / distance; // angular diameter in radians
    distance:=distance*a;           // distance in km
    parallax:=6378.137 / distance;  // horizonal parallax
    end;
  Ecl2Equ(TDT,Result);
  // Calculate horizontal coordinates of sun, if geographic positions are given
  if (not IsNan(geolat) and not IsNan(lmst)) then Equ2Altaz(TDT,geolat,lmst,Result);
  end;

// Calculate data and coordinates for the Moon
// Coordinates are accurate to about 1/5 degree (in ecliptic coordinates)
function MoonPosition (const sunCoor : TSunMoonStg; TDT : double;
                       pobserver : PSunMoonStg; lmst : double) : TSunMoonStg;
var
  p,D,xl,xi,xn,Ev,Ae,Ec,
  diameter0,parallax0,MMoon,MMoon2 : double;
const
  e = 0.054900;
  a = 384401; // km
  mainPhase = 1.0 / 29.53 * 360 * xDEG; // show 'Newmoon, 'Quarter' for +/-1 day arond the actual event
begin
  FillChar(Result,sizeof(TSunMoonStg),0);
// Mean Moon orbit elements as of 1990.0
  xi:=5.145396 * xDEG;
  diameter0:=0.5181 * xDEG; // angular diameter of Moon at a distance
  parallax0:=0.9507 * xDEG; // parallax at distance a

  D:=TDT - 2447891.5;
  xl:=13.1763966 * xDEG * D + 318.351648 * xDEG;
  MMoon:=xl - 0.1114041 * xDEG * D - 36.340410 * xDEG; // Moon's mean anomaly M
  xn:=318.510107 * xDEG - 0.0529539 * xDEG * D;        // Moon's mean ascending node longitude
  Ev:=1.2739 * xDEG * sin(2 * (xl - sunCoor.lon) - MMoon);
  Ae:=0.1858 * xDEG * sin(sunCoor.anomalyMean);
  MMoon2:=MMoon + Ev - Ae - 0.37 * xDEG * sin(sunCoor.anomalyMean);  // corrected Moon anomaly
  Ec:=6.2886 * xDEG * sin(MMoon2);  // equation of centre
  xl:=xl + Ev + Ec - Ae + 0.214 * xDEG * sin(2 * MMoon2); // corrected Moon's longitude
  xl:=xl + 0.6583 * xDEG * sin(2 * (xl - sunCoor.lon)); // true orbital longitude;
  xn:=xn - 0.16 * xDEG * sin(sunCoor.anomalyMean);
  with Result do begin
    lon:=mod2pi(xn + arctan2(sin(xl - xn) * cos(xi), cos(xl - xn)));
    lat:=arcsin(sin(xl - xn) * sin(xi));
    orbitLon:=xl;
    Ecl2Equ(TDT,Result);
    // relative distance to semi mayor axis of lunar oribt
    distance:=(1 - sqr(e)) / (1 + e * cos(MMoon2 + Ec));
    diameter:=diameter0 / distance; // angular diameter in radians
    parallax:=parallax0 / distance; // horizontal parallax in radians
    distance:=distance * a; // distance in km
    end;
  // Calculate horizonal coordinates of sun, if geographic positions is given
  if (pobserver<>nil) and not IsNan(lmst) then begin
    // transform geocentric coordinates into topocentric (==observer based) coordinates
    GeoEqu2TopoEqu(lmst,pobserver^,Result);
    with Result do begin
      raGeocentric:=ra; // backup geocentric coordinates
      decGeocentric:=dec;
      ra:=raTopocentric;
      dec:=decTopocentric;
      end;
    Equ2Altaz(TDT,pobserver^.lat,lmst,Result); // now ra and dec are topocentric
    end;
  // Age of Moon in radians since New Moon (0) - Full Moon (pi)
  with Result do begin
    moonAge:=mod2pi(xl - sunCoor.lon);
    phase:=0.5 * (1 - cos(moonAge)); // Moon phase, 0-1
    p:=FMod(moonAge, 90.0 * xDEG);
    if (p < mainPhase) or (p > 90 * xDEG - mainPhase) then moonPhase:=2 * round(moonAge / (90.0 * xDEG))
    else moonPhase:=2 * floor(moonAge / (90.0 * xDEG)) + 1;
    end;
  end;

// Rough refraction formula using standard atmosphere: 1015 mbar and 10C
// Input true altitude in radians, Output: increase in altitude in degrees
function Refraction (alt : double)  : double;
var
  i : integer;
  y,D,P,Q,y0,D0,xN : double;
const
  pressure = 1015;
  temperature = 10;
begin
  D:=alt * xRAD;
  Result:=0;
  if (D<-2) or (D >= 90) then exit;

  if (D>15) then Result:=(0.00452 * pressure / ((273 + temperature) * tan(alt)))
  else begin
    y:=alt;
    D:=0.0;
    P:=(pressure - 80.0) / 930.0;
    Q:=0.0048 * (temperature - 10.0);
    y0:=y; D0:=D;
    for i:=0 to 2 do begin
      xN:=y + (7.31 / (y + 4.4));
      xN:=1.0 / tan(xN * xDEG);
      D:=xN * P / (60.0 + Q * (xN + 39.0));
      xN:=y - y0;
      y0:=D - D0 - xN;
      if not IsZero(xN) and not IsZero(y0) then xN:=y - xN * (alt + D - y) / y0
      else xN:=alt + D;
      y0:=y; D0:=D; y:=xN;
      end;
    Result:=D;
    end;
  end;

// returns Greenwich sidereal time (hours) of time of rise
// and set of object with coordinates coor.ra/coor.dec
// at geographic position lon/lat (all values in radians)
// Correction for refraction and semi-diameter/parallax of body is taken care of in function RiseSet
// h is used to calculate the twilights. It gives the required elevation of the disk center of the sun
function GMSTRiseSet (const coor : TSunMoonStg; alon,alat,ah : double) : TSunMoonStg;
var
  tagbogen : double;
begin
  FillChar(Result,sizeof(TSunMoonStg),0);
  ah:=IfThen(IsNan(ah),0.0,ah); // set default value
  tagbogen:=(sin(ah) - sin(alat) * sin(coor.dec)) / (cos(alat) * cos(coor.dec));
  if abs(tagbogen)>1 then Result.error:=true
  else tagbogen:=arccos(tagbogen);
  with Result do if not error then begin
    transit:=xRAD / 15 * (+coor.ra - alon);
    rise:=24.0 + xRAD / 15 * (-tagbogen + coor.ra - alon); // calculate GMST of rise of object
    xset:=xRAD / 15 * (+tagbogen + coor.ra - alon); // calculate GMST of set of object
    // using the modulo function Mod, the day number goes missing. This may get a problem for the moon
    transit:=FMod(transit, 24);
    rise:=FMod(rise, 24);
    xset:=FMod(xset, 24);
    end;
  end;

// Find GMST of rise/set of object from the two calculates
// (start)points (day 1 and 2) and at midnight UT(0)
function InterpolateGMST(gmst0,gmst1,gmst2,timefactor : double) : double;
begin
  Result:=((timefactor * 24.07 * gmst1 - gmst0 * (gmst2 - gmst1)) / (timefactor * 24.07 + gmst1 - gmst2));
  end;

// JD is the Julian Date of 0h UTC time (midnight)
function RiseSet(jd0UT,alon,alat,timeinterval,naltitude : double; const coor1,coor2 : TSunMoonStg) : TSunMoonStg;
var
  rise1, rise2 : TSunMoonStg;
  T0,T02,decMean,psi,xy,dt,xalt,altitude : double;
begin
  FillChar(Result,sizeof(TSunMoonStg),0);
  // altitude of sun center: semi-diameter, horizontal parallax and (standard) refraction of 34'
  xalt:=0.0; // calculate
  altitude:=IfThen(IsNan(naltitude),0.0,naltitude); // set default value

  // true height of sun center for sunrise and set calculation. Is kept 0 for twilight (ie. altitude given):
  if IsZero(altitude) then xalt:=0.5 * coor1.diameter - coor1.parallax + 34.0 / 60 * xDEG;
  rise1:=GMSTRiseSet(coor1, alon, alat, altitude);
  rise2:=GMSTRiseSet(coor2, alon, alat, altitude);

  if rise1.error or rise2.error then with Result do begin
    error:=true;
    transit:=nan; rise:=nan; xset:=nan;
    end
  else begin
    // unwrap GMST in case we move across 24h -> 0h
    if (rise1.transit > rise2.transit) and (abs(rise1.transit - rise2.transit) > 18) then IncFloat(rise2.transit,24.0);
    if (rise1.rise > rise2.rise) and (abs(rise1.rise - rise2.rise) > 18) then IncFloat(rise2.rise,24.0);
    if (rise1.xset > rise2.xset) and (abs(rise1.xset - rise2.xset) > 18) then IncFloat(rise2.xset,24.0);
    T0:=CalcGMST(jd0UT);
    //  var T02:=T0-zone*1.002738; // Greenwich sidereal time at 0h time zone (zone: hours)

    // Greenwich sidereal time for 0h at selected longitude
    T02:=T0 - alon * xRAD / 15 * 1.002738;
    if (T02 < 0)  then IncFloat(T02,24);

    if (rise1.transit < T02) then begin
      IncFloat(rise1.transit,24.0); IncFloat(rise2.transit,24.0);
      end;
    if (rise1.rise < T02) then begin
      IncFloat(rise1.rise,24.0); IncFloat(rise2.rise,24.0);
      end;
    if (rise1.xset < T02) then begin
      IncFloat(rise1.xset,24.0); IncFloat(rise2.xset,24.0);
      end;
    // Refraction and Parallax correction
    decMean:=0.5 * (coor1.dec + coor2.dec);
    psi:=arccos(sin(alat) / cos(decMean));
    xy:=arcsin(sin(xalt) / sin(psi));
    dt:=240 * xRAD * xy / cos(decMean) / 3600; // time correction due to refraction, parallax
    with Result do begin
      transit:=gmst2ut(jd0UT,InterpolateGMST(T0, rise1.transit, rise2.transit, timeinterval));
      rise:=gmst2ut(jd0UT,InterpolateGMST(T0, rise1.rise, rise2.rise, timeinterval) - dt);
      xset:=gmst2ut(jd0UT,InterpolateGMST(T0, rise1.xset, rise2.xset, timeinterval) + dt);
      end;
    end;
  end;

// Find (local) time of sunrise and sunset, and twilights
// JD is the Julian Date of 0h local time (midnight)
// Accurate to about 1-2 minutes
// recursive: 1 - calculate rise/set in UTC in a second run
// recursive: 0 - find rise/set on the current local day. This is set when doing the first call to this function
function CalcSunRise(JD,aDeltaT,alon,alat : double; zone,recursive : integer) : TSunMoonStg;
var
  risetemp,coor1,coor2 : TSunMoonStg;
  jd0UT : double;
begin
  jd0UT:=floor(JD - 0.5) + 0.5;   // JD at 0 hours UT
  coor1:=SunPosition(jd0UT + aDeltaT / 24.0 / 3600.0, NAN, NAN);
  coor2:=SunPosition(jd0UT + 1.0 + aDeltaT / 24.0 / 3600.0, NAN, NAN); // calculations for next day's UTC midnight

  // rise/set time in UTC.
  Result:=RiseSet(jd0UT, alon, alat, 1, NAN, coor1, coor2);
  if (recursive = 0) then with Result do begin
    // check and adjust to have rise/set time on local calendar day
    if (zone > 0) then begin
      // rise time was yesterday local time -> calculate rise time for next UTC day
      if (rise >= 24 - zone) or (transit >= 24 - zone) or (xset >= 24 - zone) then begin
        risetemp:=CalcSunRise(JD + 1, aDeltaT, alon, alat, zone, 1);
        if (rise >= 24 - zone) then rise:=risetemp.rise;
        if (transit >= 24 - zone) then transit:=risetemp.transit;
        if (xset >= 24 - zone) then xset:=risetemp.xset;
        end;
      end
    else if (zone < 0) then begin
      // rise time was yesterday local time -> calculate rise time for next UTC day
      if (rise < -zone) or (transit < -zone) or (xset < -zone) then begin
        risetemp:=CalcSunRise(JD - 1, aDeltaT, alon, alat, zone, 1);
        if (rise < -zone) then rise:=risetemp.rise;
        if (transit < -zone) then transit:=risetemp.transit;
        if (xset < -zone) then xset:=risetemp.xset;
        end;
      end;
    transit:=FMod(transit + zone, 24.0);
    rise:=FMod(rise + zone, 24.0);
    xset:=FMod(xset + zone, 24.0);

    // Twilight calculation
    // civil twilight time in UTC.
    risetemp:=RiseSet(jd0UT, alon, alat, 1, -6.0 * xDEG, coor1, coor2);
    if risetemp.error then begin
      civilTwilightMorning:=nan; civilTwilightEvening:=nan;
      end
    else begin
      civilTwilightMorning:=FMod(risetemp.rise + zone, 24.0);
      civilTwilightEvening:=FMod(risetemp.xset + zone, 24.0);
      end;
    // nautical twilight time in UTC.
    risetemp:=RiseSet(jd0UT, alon, alat, 1, -12.0 * xDEG, coor1, coor2);
    if risetemp.error then begin
      nauticalTwilightMorning:=nan; nauticalTwilightEvening:=nan;
      end
    else begin
      nauticalTwilightMorning:=FMod(risetemp.rise + zone, 24.0);
      nauticalTwilightEvening:=FMod(risetemp.xset + zone, 24.0);
      end;
    // astronomical twilight time in UTC.
    risetemp:=RiseSet(jd0UT, alon, alat, 1, -18.0 * xDEG, coor1, coor2);
    if risetemp.error then begin
      astronomicalTwilightMorning:=nan; astronomicalTwilightEvening:=nan;
      end
    else begin
      astronomicalTwilightMorning:=FMod(risetemp.rise + zone, 24.0);
      astronomicalTwilightEvening:=FMod(risetemp.xset + zone, 24.0);
      end;
    end
  end;

// Find local time of moonrise and moonset
// JD is the Julian Date of 0h local time (midnight)
// Accurate to about 5 minutes or better
// recursive: 1 - calculate rise/set in UTC
// recursive: 0 - find rise/set on the current local day (set could also be first)
// returns '' for moonrise/set does not occur on selected day
function CalcMoonRise(JD,aDeltaT,alon,alat : double; zone,recursive : integer) : TSunMoonStg;
var
  risetemp,suncoor1,suncoor2,coor1,coor2 : TSunMoonStg;
  jd0UT : double;
const
  timeinterval = 0.5;
begin
  jd0UT:=floor(JD - 0.5) + 0.5;   // JD at 0 hours UT
  suncoor1:=SunPosition(jd0UT + aDeltaT / 24.0 / 3600.0, NAN, NAN);
  coor1:=MoonPosition(suncoor1, jd0UT + aDeltaT / 24.0 / 3600.0, nil, NAN);
  suncoor2:=SunPosition(jd0UT + timeinterval + aDeltaT / 24.0 / 3600.0, NAN, NAN); // calculations for noon
  // calculations for next day's midnight
  coor2:=MoonPosition(suncoor2, jd0UT + timeinterval + aDeltaT / 24.0 / 3600.0, nil, NAN);

  // rise/set time in UTC, time zone corrected later.
  // Taking into account refraction, semi-diameter and parallax
  Result:=RiseSet(jd0UT, alon, alat, timeinterval, NAN, coor1, coor2);
  if (recursive = 0) then with Result do begin
    // check and adjust to have rise/set time on local calendar day
    if (zone > 0) then begin
      // recursive call to MoonRise returns events in UTC
      risetemp:=CalcMoonRise(JD - 1.0, aDeltaT, alon, alat, zone, 1);
      if (transit >= 24.0 - zone) or (transit < -zone) then begin
        // transit time is tomorrow local time
        if (risetemp.transit < 24.0 - zone) then transit:=NAN // there is no moontransit today
        else transit:=risetemp.transit;
        end;
      if (rise >= 24.0 - zone) or (rise < -zone) then begin
        // rise time is tomorrow local time
       if (risetemp.rise < 24.0 - zone) then rise:=NAN // there is no moontransit today
       else rise:=risetemp.rise;
       end;
      if (xset >= 24.0 - zone) or (xset < -zone) then begin
        // set time is tomorrow local time
        if (risetemp.xset < 24.0 - zone) then xset:=NAN // there is no moontransit today
        else xset:=risetemp.xset;
        end;
      end
    else if (zone < 0) then begin
      // rise/set time was tomorrow local time -> calculate rise time for former UTC day
      if (rise < -zone) or (xset < -zone) or (transit < -zone) then begin
        risetemp:=CalcMoonRise(JD + 1.0, aDeltaT, alon, alat, zone, 1);
        if (rise < -zone) then begin
          if (risetemp.rise > -zone) then rise:=NAN // there is no moonrise today
          else rise:=risetemp.rise;
          end;
        if (transit < -zone) then begin
          if (risetemp.transit > -zone) then transit:=NAN // there is no moonset today
          else transit:=risetemp.transit;
          end;
        if (xset < -zone) then begin
          if (risetemp.xset > -zone) then xset:=NAN // there is no moonset today
          else xset:=risetemp.xset;
          end;
        end;
      end;
    if not IsNan(rise) then rise:=FMod(rise + zone, 24.0);    // correct for time zone, if time is valid
    if not IsNan(transit) then transit:=FMod(transit + zone, 24.0); // correct for time zone, if time is valid
    if not IsNan(xset) then xset:=FMod(xset + zone, 24.0);    // correct for time zone, if time is valid
    end;
  end;

function TimeOfDayToStr(tm : TTime; NoSeconds : boolean = false) : string;
var
  h,m,s,n : word;
begin
  if IsNan(tm) then begin
    if NoSeconds then Result:='--:--' else Result:='--:--:--';
    end
  else begin
    DecodeTime(tm/24,h,m,s,n);
    if NoSeconds then begin
      if s>=30 then inc(m);
      Result:=Format('%2.2u:%2.2u',[h,m]);
      end
    else Result:=Format('%2.2u:%2.2u:%2.2u',[h,m,s]);
    end;
  end;

function InitDeltaT : double;
var
  y : double;
begin
  // Zeitkorrektur TDT (terrestrial dynamic time) - UTC
  // 1964: 40s, 2006: 65s, 2019: 69s
  // siehe https://de.wikipedia.org/wiki/Ephemeridenzeit
  // und https://de.wikipedia.org/wiki/Delta_T
  y:=ToDay;
  y:=YearOf(y)+(MonthOf(y)-0.5)/12-2000;
  Result:=62.92+(0.005589*y+0.32217)*y;
  end;

function InitData : TSunMoonData;
begin
  FillChar(Result,sizeof(TSunMoonData),0);
  with Result do begin
    Lat:=50; Lon:=10; DeltaT:=InitDeltaT;
    end;
  end;

function ComputeSunMoon (latitude,longitude,height : double; tm : TDateTime; aDeltaT : double; aZone : integer) : TSunMoonData;
var
  observerCart, sunCoor, moonCoor, sunCart, sunTimes, moonCart, moonTimes : TSunMoonStg;
  JD0,jd1,TDT,xlat,xlon,fgmst,flmst : double;
begin
  JD0:=CalcJD(DayOf(tm),MonthOf(tm),YearOf(tm));
  jd1:=JD0 + (HourOf(tm)-aZone + MinuteOf(tm)/60.0 + SecondOf(tm)/3600.0) / 24.0;
  xlat:=latitude * xDEG;  // geodetic latitude of observer on WGS84
  xlon:=longitude * xDEG; // latitude of observer
  // altitude of observer in meters above WGS84 ellipsoid (and converted to kilometers)
  // see: https://de.wikipedia.org/wiki/H%C3%B6he_(Geod%C3%A4sie)#Ellipsoidische_H%C3%B6hen
  height:=height * 0.001;
  fgmst:=CalcGMST(jd1);
  flmst:=GMST2LMST(fgmst, xlon);
  with Result do begin
    Lat:=latitude; Lon:=longitude; DateTime:=tm;
    if IsNan(aDeltaT) then DeltaT:=InitDeltaT  else DeltaT:=aDeltaT;
    TDT:=jd1 + DeltaT / 24.0 / 3600.0;
    Zone:=aZone;
    JD:=RoundTo(jd1,-5);
    GMST:=fgmst;
    LMST:=flmst;
    EOT:=calcEOT(jd1);
    end;
  observerCart:=Observer2EquCart(xlon, xlat, height, fgmst); // geocentric cartesian coordinates of observer
  sunCoor:=SunPosition(TDT, xlat, flmst * 15.0 *xDEG);   // Calculate data for the Sun at given time
  moonCoor:=MoonPosition(sunCoor, TDT, @observerCart, flmst * 15.0 * xDEG);    // Calculate data for the Moon at given time
  with Result do begin
    SunLon:=RoundTo(sunCoor.lon * xRAD,-3);
    SunRA:=sunCoor.ra * xRAD / 15;
    SunDec:=RoundTo(sunCoor.dec * xRAD,-3);
    SunAz:=RoundTo(sunCoor.az * xRAD,-2);
    SunAlt:=RoundTo(sunCoor.alt * xRAD + Refraction(sunCoor.alt),-2);  // including refraction

    SunSign:=TStarSign(trunc(sunCoor.lon * xRAD / 30.0));
    SunDiameter:=RoundTo(sunCoor.diameter * xRAD * 60.0,-2); // angular diameter in arc seconds
    SunDistance:=RoundTo(sunCoor.distance,-1);

    // Calculate distance from the observer (on the surface of earth) to the center of the sun
    sunCart:=EquPolar2Cart(sunCoor.ra, sunCoor.dec, sunCoor.distance);
    SunDistanceObserver:=RoundTo(sqrt(sqr(sunCart.x - observerCart.x) + sqr(sunCart.y - observerCart.y) + sqr(sunCart.z - observerCart.z)),-1);

    // JD0: JD of 0h UTC time
    sunTimes:=CalcSunRise(JD0, DeltaT, xlon, xlat, aZone, 0);

    SunTransit:=sunTimes.transit;
    SunRise:=sunTimes.rise;
    SunSet:=sunTimes.xset;

    SunCivilTwilightMorning:=sunTimes.civilTwilightMorning;
    SunCivilTwilightEvening:=sunTimes.civilTwilightEvening;
    SunNauticalTwilightMorning:=sunTimes.nauticalTwilightMorning;
    SunNauticalTwilightEvening:=sunTimes.nauticalTwilightEvening;
    SunAstronomicalTwilightMorning:=sunTimes.astronomicalTwilightMorning;
    SunAstronomicalTwilightEvening:=sunTimes.astronomicalTwilightEvening;

    MoonLon:=RoundTo(moonCoor.lon * xRAD,-3);
    MoonLat:=RoundTo(moonCoor.lat * xRAD,-3);
    MoonRA:=moonCoor.ra * xRAD / 15.0;
    MoonDec:=RoundTo(moonCoor.dec * xRAD,-3);
    MoonAz:=RoundTo(moonCoor.az * xRAD,-2);
    MoonAlt:=RoundTo(moonCoor.alt * xRAD + Refraction(moonCoor.alt),-1);  // including refraction
    MoonAge:=RoundTo(moonCoor.moonAge * xRAD,-3);
    MoonPhaseNumber:=RoundTo(moonCoor.phase,-3);
    if (moonCoor.moonPhase = 8) then moonCoor.moonPhase:=0;
    MoonPhase:=TLunarPhase(moonCoor.moonPhase);

    MoonSign:=TStarSign(trunc(moonCoor.lon * xRAD / 30.0));
    MoonDistance:=RoundTo(moonCoor.distance,-1);
    MoonDiameter:=RoundTo(moonCoor.diameter * xRAD * 60.0,-2); // angular diameter in arc seconds

    // Calculate distance from the observer (on the surface of earth) to the center of the moon
    moonCart:=EquPolar2Cart(moonCoor.raGeocentric, moonCoor.decGeocentric, moonCoor.distance);
    MoonDistanceObserver:=RoundTo(sqrt(sqr(moonCart.x - observerCart.x) + sqr(moonCart.y - observerCart.y) + sqr(moonCart.z - observerCart.z)),-1);

    moonTimes:=CalcMoonRise(JD0, DeltaT, xlon, xlat, aZone, 0);

    MoonTransit:=moonTimes.transit;
    MoonRise:=moonTimes.rise;
    MoonSet:=moonTimes.xset;
    end;
  end;

function ComputeSunTimes  (latitude,longitude : double; tm : TDateTime; aDeltaT : double; aZone : integer) : TSunTimes;
var
  sunTimes : TSunMoonStg;
  JD0,xlat,xlon : double;
begin
  JD0:=CalcJD(DayOf(tm),MonthOf(tm),YearOf(tm));
  xlat:=latitude * xDEG;  // geodetic latitude of observer on WGS84
  xlon:=longitude * xDEG; // latitude of observer
  // altitude of observer in meters above WGS84 ellipsoid (and converted to kilometers)
  // see: https://de.wikipedia.org/wiki/H%C3%B6he_(Geod%C3%A4sie)#Ellipsoidische_H%C3%B6hen
  if IsNan(aDeltaT) then aDeltaT:=initDeltaT;
  with Result do begin
    // JD0: JD of 0h UTC time
    sunTimes:=CalcSunRise(JD0, aDeltaT, xlon, xlat, aZone, 0);
    SunTransit:=sunTimes.transit;
    SunRise:=sunTimes.rise;
    SunSet:=sunTimes.xset;
    end;
  end;

function SunRightAscension (latitude,longitude : double; tm : TDateTime; aDeltaT : double; aZone : integer) : TTime;
var
  JD,flmst : double;
begin
  JD:=CalcJD(DayOf(tm),MonthOf(tm),YearOf(tm))
    +(HourOf(tm)-aZone + MinuteOf(tm)/60.0 + SecondOf(tm)/3600.0)/24.0;
  if IsNan(aDeltaT) then aDeltaT:=initDeltaT;
  Result:=SunPosition(JD+aDeltaT/24.0/3600.0,latitude*xDEG,
    GMST2LMST(CalcGMST(JD),longitude*xDEG)*15.0*xDEG).ra*xRAD/15;
  end;

function MoonPhase (latitude,longitude : double; tm : TDateTime; aDeltaT : double; aZone : integer) : TLunarPhase;
var
  JD,TDT,lmst : double;
begin
  if IsNan(aDeltaT) then aDeltaT:=initDeltaT;
  JD:=CalcJD(DayOf(tm),MonthOf(tm),YearOf(tm))
    +(HourOf(tm)-aZone + MinuteOf(tm)/60.0 + SecondOf(tm)/3600.0)/24.0;
  TDT:=JD+aDeltaT/24.0/3600.0;
  lmst:=GMST2LMST(CalcGMST(JD),longitude*xDEG)*15.0*xDEG;
  Result:=TLunarPhase(MoonPosition(SunPosition(TDT,latitude*xDEG,lmst),TDT,nil,lmst).moonPhase);
  end;

function MoonPhaseNumber (latitude,longitude : double; tm : TDateTime; aDeltaT : double; aZone : integer) : double;
var
  JD,TDT,lmst : double;
begin
  if IsNan(aDeltaT) then aDeltaT:=initDeltaT;
  JD:=CalcJD(DayOf(tm),MonthOf(tm),YearOf(tm))
    +(HourOf(tm)-aZone + MinuteOf(tm)/60.0 + SecondOf(tm)/3600.0)/24.0;
  TDT:=JD+aDeltaT/24.0/3600.0;
  lmst:=GMST2LMST(CalcGMST(JD),longitude*xDEG)*15.0*xDEG;
  Result:=MoonPosition(SunPosition(TDT,latitude*xDEG,lmst),TDT,nil,lmst).phase;
  end;

function MoonAge (latitude,longitude : double; tm : TDateTime; aDeltaT : double; aZone : integer) : double;
var
  JD,TDT,lmst : double;
begin
  if IsNan(aDeltaT) then aDeltaT:=initDeltaT;
  JD:=CalcJD(DayOf(tm),MonthOf(tm),YearOf(tm))
    +(HourOf(tm)-aZone + MinuteOf(tm)/60.0 + SecondOf(tm)/3600.0)/24.0;
  TDT:=JD+aDeltaT/24.0/3600.0;
  lmst:=GMST2LMST(CalcGMST(JD),longitude*xDEG)*15.0*xDEG;
  Result:=MoonPosition(SunPosition(TDT,latitude*xDEG,lmst),TDT,nil,lmst).moonAge*xRAD;
  end;

end.

